"""Rabbit Air TCP-based client."""

import asyncio
import socket
import struct
from typing import Any, Dict

from .client import Client
from .exceptions import NetworkError


class TcpClient(Client):
    """TCP-based client."""

    timeout: float = 5.0

    @classmethod
    def _create_socket(cls) -> socket.socket:
        return socket.socket(type=socket.SOCK_STREAM)

    @staticmethod
    async def _recvall(sock: socket.socket, size: int) -> bytes:
        data = b""
        loop = asyncio.get_running_loop()
        while len(data) < size:
            chunk = await loop.sock_recv(sock, size - len(data))
            if not chunk:
                break
            data += chunk
        if len(data) != size:
            raise NetworkError("Connection was unexpectedly closed")
        return data

    async def _recvmsg(self) -> bytes:
        assert self._sock is not None
        header = await self._recvall(self._sock, 2)
        size = struct.unpack("<H", header)[0]
        return await self._recvall(self._sock, size)

    async def _sendmsg(self, data: bytes) -> None:
        assert self._sock is not None
        loop = asyncio.get_running_loop()
        await loop.sock_sendall(self._sock, struct.pack("<H", len(data)) + data)

    async def _exchange(self, request_id: int, data: bytes) -> Dict[str, Any]:
        return await asyncio.wait_for(super()._exchange(request_id, data), self.timeout)



"""Rabbit Air Python library exceptions."""


class NetworkError(Exception):
    """The network connection behaved in an unexpected way."""

    pass


class ProtocolError(Exception):
    """The protocol message was recognized but has invalid parameters."""

    pass



"""Rabbit Air protocol client."""

import asyncio
import json
import os
import socket
import time
from abc import ABC, abstractmethod
from random import SystemRandom
from types import TracebackType
from typing import Any, Dict, List, Optional, Type

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from zeroconf.asyncio import AsyncZeroconf

from .exceptions import ProtocolError
from .response import (
    Info,
    Lights,
    Mode,
    Moodlight,
    Sensitivity,
    Speed,
    State,
    TimerMode,
)


class Client(ABC):
    """Base class for the Rabbit Air protocol client.

    To create an instance of the class use UdpClient or TcpClient.
    """

    _sock: Optional[socket.socket] = None
    _ts_diff: Optional[float] = None

    def __init__(
        self,
        host: str,
        token: Optional[str],
        port: int = 9009,
        zeroconf: Optional[AsyncZeroconf] = None,
    ) -> None:
        """Initialize the client."""
        self._host = host
        self._token = bytes.fromhex(token) if token else None
        if self._token and len(self._token) != 16:
            raise ValueError("Invalid token length")
        self._port = port
        self._zeroconf = zeroconf
        self._id = SystemRandom().randrange(0x1000000)
        self._lock = asyncio.Lock()

    def __enter__(self) -> "Client":
        """Enter the runtime context related to this object."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Close the client."""
        if self._sock is not None:
            self._sock.close()
        return None

    @classmethod
    @abstractmethod
    def _create_socket(cls) -> socket.socket:
        """Create a network socket."""

    async def _resolve(self, host: str) -> str:
        if self._zeroconf is not None and host.endswith(".local"):
            info = await self._zeroconf.async_get_service_info(
                "_rabbitair._udp.local.", host + "."
            )
            if info:
                addr = info.parsed_addresses()
                if len(addr) > 0:
                    return addr[0]
        return host

    async def _start(self) -> None:
        assert self._sock is None
        self._sock = self._create_socket()
        self._sock.setblocking(False)
        loop = asyncio.get_running_loop()
        host = await self._resolve(self._host)
        await loop.sock_connect(self._sock, (host, self._port))

    def _stop(self) -> None:
        assert self._sock is not None
        self._sock.close()
        self._sock = None
        self._ts_diff = None

    @staticmethod
    def _clock() -> float:
        return time.clock_gettime(time.CLOCK_BOOTTIME)

    def _get_ts(self) -> int:
        assert self._ts_diff is not None
        return round(self._clock() + self._ts_diff)

    def _next_id(self) -> int:
        value = self._id
        self._id += 1
        return value

    def _decrypt(self, msg: bytes) -> bytes:
        assert self._token is not None
        iv = msg[-16:]
        msg = msg[:-16]
        cipher = Cipher(algorithms.AES(self._token), modes.CBC(iv), default_backend())
        decryptor = cipher.decryptor()  # type: ignore[no-untyped-call]
        msg = decryptor.update(msg) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()  # type: ignore[no-untyped-call]
        return unpadder.update(msg) + unpadder.finalize()  # type: ignore[no-any-return]

    def _encrypt(self, msg: bytes) -> bytes:
        assert self._token is not None
        padder = padding.PKCS7(128).padder()  # type: ignore[no-untyped-call]
        msg = padder.update(msg) + padder.finalize()
        iv = os.urandom(16)
        cipher = Cipher(algorithms.AES(self._token), modes.CBC(iv), default_backend())
        encryptor = cipher.encryptor()  # type: ignore[no-untyped-call]
        return encryptor.update(msg) + encryptor.finalize() + iv  # type: ignore[no-any-return]

    @abstractmethod
    async def _recvmsg(self) -> bytes:
        """Receive messages over the network."""

    @abstractmethod
    async def _sendmsg(self, data: bytes) -> None:
        """Send messages over the network."""

    async def _exchange(self, request_id: int, data: bytes) -> Dict[str, Any]:
        await self._sendmsg(data)
        while True:
            data = await self._recvmsg()
            if self._token:
                data = self._decrypt(data)
            try:
                response: Dict[str, Any] = json.loads(data)
                if response["id"] == request_id:
                    return response
            except (json.JSONDecodeError, KeyError):
                # Ignore any garbage or unexpected responses
                pass

    async def _command(self, request: Dict[str, Any]) -> Dict[str, Any]:
        data = json.dumps(request, separators=(",", ":")).encode()
        if self._token:
            data = self._encrypt(data)
        response = await self._exchange(request["id"], data)
        if response.get("error"):
            raise ProtocolError()
        return response

    async def command(self, request: Dict[str, Any]) -> Dict[str, Any]:
        """Send raw command to the device."""
        async with self._lock:
            try:
                if not self._sock:
                    await self._start()
                if self._token:
                    if self._ts_diff is None:
                        ts_request = {"id": self._next_id(), "cmd": 9}
                        response = await self._command(ts_request)
                        ts = response["data"]["ts"]
                        self._ts_diff = ts - self._clock()
                        request["ts"] = ts
                    else:
                        request["ts"] = self._get_ts()
                request["id"] = self._next_id()
                return await self._command(request)
            except ProtocolError:
                raise
            except Exception:
                self._stop()
                raise

    async def get_state(self) -> State:
        """Get the current state of the device."""
        response = await self.command({"cmd": 4})
        return State(response["data"])

    async def set_state(
        self,
        power: Optional[bool] = None,
        mode: Optional[Mode] = None,
        speed: Optional[Speed] = None,
        sensitivity: Optional[Sensitivity] = None,
        ionizer: Optional[bool] = None,
        moodlight: Optional[Moodlight] = None,
        filter_cleaning: Optional[bool] = None,
        filter_replacement: Optional[bool] = None,
        filter_life: Optional[int] = None,
        filter_timer: Optional[int] = None,
        lights: Optional[Lights] = None,
        color: Optional[List[int]] = None,
        light_sensor_ctl: Optional[bool] = None,
        filter_ctl: Optional[bool] = None,
        buzzer: Optional[bool] = None,
        child_lock: Optional[bool] = None,
        timer_mode: Optional[TimerMode] = None,
        timer: Optional[int] = None,
        schedule: Optional[str] = None,
    ) -> None:
        """Change the state of the device."""
        data: Dict[str, Any] = {}
        if power is not None:
            data["power"] = power
        if mode is not None:
            data["mode"] = mode.value
        if speed is not None:
            data["speed"] = speed.value
        if sensitivity is not None:
            data["sensitivity"] = sensitivity.value
        if ionizer is not None:
            data["ionizer"] = ionizer
        if moodlight is not None:
            data["moodlight"] = moodlight.value
        if filter_cleaning is not None:
            data["filter_cleaning"] = filter_cleaning
        if filter_replacement is not None:
            data["filter_replacement"] = filter_replacement
        if filter_life is not None:
            if filter_life < 0 or filter_life > 525600:
                raise ValueError("The filter life value must be in the range 0-1440")
            data["filter_life"] = filter_life
        if filter_timer is not None:
            if filter_timer < 0 or filter_timer > 525600:
                raise ValueError("The filter timer value must be in the range 0-1440")
            data["filter_timer"] = filter_timer
        if lights is not None:
            data["all_light_off"] = lights.value
        if color is not None:
            if len(color) != 9:
                raise ValueError("The color length must be 9")
            for v in color:
                if v < 0 or v > 40:
                    raise ValueError("The color values must be in the range 0-40")
            data["color"] = color
        if light_sensor_ctl is not None:
            data["lsens_ctl"] = light_sensor_ctl
        if filter_ctl is not None:
            data["filter_ctl"] = filter_ctl
        if buzzer is not None:
            data["buzzer"] = buzzer
        if child_lock is not None:
            data["lock"] = child_lock
        if timer_mode is not None:
            data["timer_mode"] = timer_mode.value
        if timer is not None:
            if timer < 0 or timer > 1440:
                raise ValueError("The timer value must be in the range 0-1440")
            data["timer"] = timer
        if schedule is not None:
            if len(schedule) != 24:
                raise ValueError("The schedule length must be 24")
            for c in schedule:
                if (c < "0" or c > "5") and c != "A":
                    raise ValueError("The schedule values must be 0-5,A")
            data["schedule"] = schedule
        await self.command({"cmd": 4, "data": data})

    async def get_info(self) -> Info:
        """Get information about the device."""
        response = await self.command({"cmd": 255})
        return Info(response["data"])



"""Rabbit Air Python library."""

from .client import Client
from .exceptions import NetworkError, ProtocolError
from .response import (
    RSSI,
    Error,
    FilterType,
    Gas,
    Info,
    Lights,
    Mode,
    Model,
    Moodlight,
    Quality,
    Sensitivity,
    Speed,
    State,
    TimerMode,
)
from .tcp import TcpClient
from .udp import UdpClient

__all__ = [
    "Client",
    "Error",
    "FilterType",
    "Gas",
    "Info",
    "Lights",
    "Mode",
    "Model",
    "Moodlight",
    "NetworkError",
    "ProtocolError",
    "Quality",
    "RSSI",
    "Sensitivity",
    "Speed",
    "State",
    "TcpClient",
    "TimerMode",
    "UdpClient",
]



"""Rabbit Air UDP-based client."""

import asyncio
import socket
from typing import Any, Dict

from .client import Client


class UdpClient(Client):
    """UDP-based client."""

    retry_count: int = 3
    timeout: float = 2.0

    @classmethod
    def _create_socket(cls) -> socket.socket:
        return socket.socket(type=socket.SOCK_DGRAM)

    async def _recvmsg(self) -> bytes:
        assert self._sock is not None
        loop = asyncio.get_running_loop()
        return await loop.sock_recv(self._sock, 2048)

    async def _sendmsg(self, data: bytes) -> None:
        assert self._sock is not None
        loop = asyncio.get_running_loop()
        await loop.sock_sendall(self._sock, data)

    async def _exchange(self, request_id: int, data: bytes) -> Dict[str, Any]:
        for i in range(self.retry_count):
            try:
                return await asyncio.wait_for(
                    super()._exchange(request_id, data), self.timeout
                )
            except asyncio.TimeoutError as e:
                error = e
        raise error



"""Rabbit Air responses."""

import inspect
from enum import Enum, unique
from typing import Generic, List, Optional, TypeVar

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

T = TypeVar("T")


@unique
class Model(Enum):
    """Device model."""

    MinusA2 = 1
    BioGS = 2
    A3 = 3


@unique
class Mode(Enum):
    """Mode of operation."""

    Auto = 0
    Pollen = 1
    Manual = 2


@unique
class Speed(Enum):
    """Fan speed.

    Note: SuperSilent mode cannot be set manually.
    """

    SuperSilent = 0
    Silent = 1
    Low = 2
    Medium = 3
    High = 4
    Turbo = 5


@unique
class Quality(Enum):
    """Air quality.

    Note: the Lowest value is not actually used.
    """

    Lowest = 0
    Low = 1
    Medium = 2
    High = 3
    Highest = 4


@unique
class Sensitivity(Enum):
    """Sensitivity level of the sensors."""

    High = 0
    Medium = 1
    Low = 2


class Moodlight(Enum):
    """Mood Light mode."""

    Off = 0
    On = 1
    Auto = 2
    Preset1 = 2
    Preset2 = 3
    Preset3 = 4


@unique
class Lights(Enum):
    """Light settings."""

    Off = 0
    On = 1
    Auto = 2


@unique
class Error(Enum):
    """Internal error codes."""

    NoError = 0
    DustSensor = 1
    GasSensor = 2
    GasAndDust = 3
    FanLow = 4
    NFC = 5
    HallSensor = 8


@unique
class FilterType(Enum):
    """Filter type."""

    Unknown = 0
    ToxinAbsorber = 1
    OdorRemover = 2
    GermDefense = 3
    PetAllergy = 4


@unique
class Gas(Enum):
    """Gas sensor readings."""

    Preheat = 0
    Level1 = 1
    Level2 = 2
    Level3 = 3
    Level4 = 4


@unique
class TimerMode(Enum):
    """Timer mode."""

    Off = 0
    On = 1
    Schedule = 2


class StateDict(TypedDict, total=False):
    """State response typing."""

    model: int
    firmware: List[int]
    power: bool
    mode: int
    speed: int
    quality: int
    sensitivity: int
    ionizer: bool
    idle: int
    moodlight: int
    sleep: bool
    filter_cleaning: bool
    filter_replacement: bool
    filter_life: int
    light_sensor: bool
    particulate_sensor: int
    filter_timer: int
    all_light_off: int
    error: int
    tag_state: int
    tag_uid: List[int]
    filter_type: int
    pm_sensor: List[int]
    color: List[int]
    lsens_ctl: bool
    filter_ctl: bool
    buzzer: bool
    gas: int
    lock: bool
    open: bool
    timer_mode: int
    timer: int
    schedule: str
    rssi: int
    v: str


class RssiDict(TypedDict, total=False):
    """RSSI typing."""

    cur: int
    min: int
    max: int
    avg: int


class InfoDict(TypedDict, total=False):
    """Info response typing."""

    name: str
    mcu: str
    build: str
    mac: str
    time: str
    uptime: int
    mup: int
    wup: int
    iup: int
    cup: int
    fv: str
    rssi: RssiDict


class Response(Generic[T]):
    """Base class for the device response."""

    def __init__(self, data: T) -> None:
        """Initialize."""
        self.data = data

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        items = []
        props = inspect.getmembers(type(self), lambda x: isinstance(x, property))
        for name, prop in props:
            try:
                value = prop.fget(self)
            except Exception as ex:
                value = repr(ex)
            items.append(f"{name}={value}")
        return f"<{type(self).__name__} {' '.join(items)}>"


class State(Response[StateDict]):
    """Device state."""

    @property
    def model(self) -> Optional[Model]:
        """Device model."""
        value = self.data.get("model")
        return None if value is None else Model(value)

    @property
    def main_firmware(self) -> Optional[str]:
        """Version of the main board firmware."""
        value = self.data.get("firmware")
        if value is None:
            return None
        return ".".join(str(x) for x in value)

    @property
    def power(self) -> Optional[bool]:
        """Power on/off."""
        return self.data.get("power")

    @property
    def mode(self) -> Optional[Mode]:
        """Mode of operation."""
        value = self.data.get("mode")
        return None if value is None else Mode(value)

    @property
    def speed(self) -> Optional[Speed]:
        """Fan speed."""
        value = self.data.get("speed")
        return None if value is None else Speed(value)

    @property
    def quality(self) -> Optional[Quality]:
        """Air quality."""
        value = self.data.get("quality")
        if value is None:
            return None
        if self.model is Model.BioGS:
            return Quality(value - 1)
        else:
            return Quality(value)

    @property
    def sensitivity(self) -> Optional[Sensitivity]:
        """Sensitivity level of the sensors."""
        value = self.data.get("sensitivity")
        return None if value is None else Sensitivity(value)

    @property
    def ionizer(self) -> Optional[bool]:
        """Negative ion on/off."""
        return self.data.get("ionizer")

    @property
    def idle(self) -> Optional[bool]:
        """Device is in idle mode."""
        value = self.data.get("idle")
        return None if value is None else bool(value)

    @property
    def moodlight(self) -> Optional[Moodlight]:
        """Mood Light mode."""
        value = self.data.get("moodlight")
        return None if value is None else Moodlight(value)

    @property
    def sleep(self) -> Optional[bool]:
        """Device is in sleep mode."""
        return self.data.get("sleep")

    @property
    def filter_cleaning(self) -> Optional[bool]:
        """Filter cleaning required."""
        return self.data.get("filter_cleaning")

    @property
    def filter_replacement(self) -> Optional[bool]:
        """Filter replacement required."""
        return self.data.get("filter_replacement")

    @property
    def filter_life(self) -> Optional[int]:
        """Remaining filter lifetime."""
        return self.data.get("filter_life")

    @property
    def light_sensor(self) -> Optional[bool]:
        """Light sensor readings."""
        return self.data.get("light_sensor")

    @property
    def particulate_sensor(self) -> Optional[int]:
        """Particle sensor readings."""
        return self.data.get("particulate_sensor")

    @property
    def filter_timer(self) -> Optional[int]:
        """Nominal filter lifetime."""
        return self.data.get("filter_timer")

    @property
    def lights(self) -> Optional[Lights]:
        """Turn all light on/off."""
        value = self.data.get("all_light_off")
        return None if value is None else Lights(value)

    @property
    def error(self) -> Optional[Error]:
        """Internal error codes."""  # noqa: D401
        value = self.data.get("error")
        return None if value is None else Error(value)

    @property
    def tag_state(self) -> Optional[bool]:
        """Filter tag state."""
        value = self.data.get("tag_state")
        return None if value is None else bool(value)

    @property
    def tag_uid(self) -> Optional[List[int]]:
        """Filter tag unique identifier."""
        return self.data.get("tag_uid")

    @property
    def filter_type(self) -> Optional[FilterType]:
        """Filter type."""
        value = self.data.get("filter_type")
        return None if value is None else FilterType(value)

    @property
    def pm_sensor(self) -> Optional[List[int]]:
        """Extended particle sensor readings."""  # noqa: D401
        return self.data.get("pm_sensor")

    @property
    def color(self) -> Optional[List[int]]:
        """Color palette for Mood Light."""
        return self.data.get("color")

    @property
    def light_sensor_ctl(self) -> Optional[bool]:
        """Activate/deactivate light sensor."""
        return self.data.get("lsens_ctl")

    @property
    def filter_ctl(self) -> Optional[bool]:
        """Activate/deactivate notification about filter replacement condition."""
        return self.data.get("filter_ctl")

    @property
    def buzzer(self) -> Optional[bool]:
        """Buzzer sound on/off."""
        return self.data.get("buzzer")

    @property
    def gas(self) -> Optional[Gas]:
        """Gas sensor readings."""
        value = self.data.get("gas")
        return None if value is None else Gas(value)

    @property
    def child_lock(self) -> Optional[bool]:
        """Child lock on/off."""
        return self.data.get("lock")

    @property
    def open(self) -> Optional[bool]:
        """Front panel is removed or open."""
        return self.data.get("open")

    @property
    def timer_mode(self) -> Optional[TimerMode]:
        """Timer mode."""
        value = self.data.get("timer_mode")
        return None if value is None else TimerMode(value)

    @property
    def timer(self) -> Optional[int]:
        """Time, in minutes, remaining until the unit is turned off."""
        return self.data.get("timer")

    @property
    def schedule(self) -> Optional[str]:
        """24-h schedule.

        This is a 24-character string in which each character specifies the speed for a specific hour.
        Acceptable values are 1-5 and A (auto). The time is in UTC.
        """
        return self.data.get("schedule")

    @property
    def rssi(self) -> Optional[int]:
        """Wi-Fi RSSI value averaged over an hour."""
        return self.data.get("rssi")

    @property
    def wifi_firmware(self) -> Optional[str]:
        """Version of the Wi-Fi board firmware."""
        return self.data.get("v")


class RSSI(Response[RssiDict]):
    """Detailed information about Wi-Fi RSSI."""

    @property
    def current(self) -> Optional[int]:
        """Current RSSI value."""  # noqa: D401
        return self.data.get("cur")

    @property
    def min(self) -> Optional[int]:
        """Minimal RSSI value for an hour."""
        return self.data.get("min")

    @property
    def max(self) -> Optional[int]:
        """Maximal RSSI value for an hour."""
        return self.data.get("max")

    @property
    def average(self) -> Optional[int]:
        """RSSI value averaged over an hour."""
        return self.data.get("avg")


class Info(Response[InfoDict]):
    """Information about the device."""

    @property
    def name(self) -> str:
        """Device ID."""
        return self.data["name"]

    @property
    def wifi_firmware(self) -> str:
        """Version of the Wi-Fi board firmware."""
        return self.data["mcu"]

    @property
    def build(self) -> str:
        """Build date of the Wi-Fi board firmware."""
        return self.data["build"]

    @property
    def mac(self) -> str:
        """MAC address."""
        return self.data["mac"]

    @property
    def time(self) -> Optional[str]:
        """Current time."""  # noqa: D401
        return self.data.get("time")

    @property
    def uptime(self) -> int:
        """Total uptime."""
        return self.data["uptime"]

    @property
    def motor_uptime(self) -> int:
        """Total motor runtime."""
        return self.data["mup"]

    @property
    def wifi_uptime(self) -> int:
        """Total Wi-Fi uptime."""
        return self.data["wup"]

    @property
    def internet_uptime(self) -> Optional[int]:
        """Total internet uptime."""
        return self.data.get("iup")

    @property
    def cloud_uptime(self) -> Optional[int]:
        """Total cloud uptime."""
        return self.data.get("cup")

    @property
    def main_firmware(self) -> Optional[str]:
        """Version of the main board firmware."""
        return self.data.get("fv")

    @property
    def rssi(self) -> Optional[RSSI]:
        """Detailed information about Wi-Fi RSSI."""
        value = self.data.get("rssi")
        return None if value is None else RSSI(value)
